@@ -5,16 +5,14 @@
 #include <linux/uinput.h>
 
 #include "rf-common.h"
-#include "rf-buffer.h"
 #include "rf-streamer.h"
 
 #define KEYBOARD_MAX_EVENTS 2
 #define POINTER_MAX_EVENTS 10
 
 struct _RfStreamer {
-	GObject parent_instance;
+	GSocketClient parent_instance;
 	RfConfig *config;
-	GSocketClient *client;
 	GSocketAddress *address;
 	GSocketConnection *connection;
 	GIOCondition io_flags;
@@ -32,7 +30,7 @@ struct _RfStreamer {
 	uint32_t frame_height;
 	bool running;
 };
-G_DEFINE_TYPE(RfStreamer, rf_streamer, G_TYPE_OBJECT)
+G_DEFINE_TYPE(RfStreamer, rf_streamer, G_TYPE_SOCKET_CLIENT)
 
 enum {
 	SIG_START,
@@ -40,11 +38,36 @@ enum {
 	SIG_FRAME,
 	SIG_CARD_PATH,
 	SIG_CONNECTOR_NAME,
+	SIG_AUTH,
 	N_SIGS
 };
 
 static unsigned int sigs[N_SIGS] = { 0 };
 
+static void _send_auth_msg(RfStreamer *this, pid_t pid)
+{
+	ssize_t ret = 0;
+	g_autoptr(GError) error = NULL;
+	GOutputStream *os =
+		g_io_stream_get_output_stream(G_IO_STREAM(this->connection));
+
+	ret = rf_send_header(this->connection, RF_MSG_TYPE_AUTH, 1, &error);
+	if (ret <= 0)
+		goto out;
+	ret = g_output_stream_write(os, &pid, sizeof(pid), NULL, &error);
+
+out:
+	if (ret < 0) {
+		g_warning("Auth: Failed to send auth PID: %s.", error->message);
+		rf_streamer_stop(this);
+	} else if (ret == 0) {
+		g_warning("ReFrame Streamer disconnected.");
+		rf_streamer_stop(this);
+	} else {
+		g_debug("Auth: Sent auth message for PID %d.", pid);
+	}
+}
+
 static void
 _send_input_msg(RfStreamer *this, struct input_event *ies, const size_t length)
 {
@@ -79,7 +102,7 @@ out:
 	}
 }
 
-static gboolean _send_frame_msg(gpointer data)
+static int _send_frame_msg(void *data)
 {
 	RfStreamer *this = data;
 	ssize_t ret = 0;
@@ -106,8 +129,8 @@ static void _schedule_frame_msg(RfStreamer *this)
 	if (this->timer_id != 0)
 		return;
 
-	int64_t current = g_get_monotonic_time();
-	int64_t delta = current - this->last_frame_time;
+	const int64_t current = g_get_monotonic_time();
+	const int64_t delta = current - this->last_frame_time;
 	// Give this highest priority to prevent lag.
 	if (delta < this->max_interval) {
 		this->timer_id = g_timeout_add_full(
@@ -131,7 +154,7 @@ static void _schedule_frame_msg(RfStreamer *this)
 }
 
 static ssize_t
-_on_buffer(GSocketConnection *connection, RfBuffer *b, GError **error)
+_on_buffer(GSocketConnection *connection, struct rf_buffer *b, GError **error)
 {
 	ssize_t ret = 0;
 	GSocket *socket = g_socket_connection_get_socket(connection);
@@ -180,7 +203,7 @@ _on_buffer(GSocketConnection *connection, RfBuffer *b, GError **error)
 	// See <https://docs.gtk.org/gio-unix/type_func.FDMessage.get_fd_list.html>.
 	GUnixFDMessage *msg = G_UNIX_FD_MESSAGE(msgs[n_msgs - 1]);
 	GUnixFDList *fds = g_unix_fd_message_get_fd_list(msg);
-	for (int i = 0; i < b->md.length; ++i) {
+	for (unsigned int i = 0; i < b->md.length; ++i) {
 		b->fds[i] = g_unix_fd_list_get(fds, i, NULL);
 		// Some error happens.
 		if (b->fds[i] == -1) {
@@ -201,7 +224,7 @@ _on_buffer(GSocketConnection *connection, RfBuffer *b, GError **error)
 
 out:
 	for (int i = 0; i < n_msgs; ++i)
-		g_object_unref(msgs[i]);
+		g_clear_object(&msgs[i]);
 
 	return ret;
 }
@@ -210,9 +233,9 @@ static ssize_t _on_frame_msg(RfStreamer *this)
 {
 	g_debug("Frame: Received frame message.");
 
-	RfBuffer bufs[RF_MAX_BUFS];
-	size_t length = 0;
+	struct rf_buffer bufs[RF_MAX_BUFS];
 	ssize_t ret = 0;
+	size_t length = 0;
 	g_autoptr(GError) error = NULL;
 	GInputStream *is =
 		g_io_stream_get_input_stream(G_IO_STREAM(this->connection));
@@ -237,7 +260,7 @@ static ssize_t _on_frame_msg(RfStreamer *this)
 			goto out;
 	}
 
-	RfBuffer *primary = &bufs[0];
+	struct rf_buffer *primary = &bufs[0];
 	// Monitor size should be CRTC size, and primary plane is used to store
 	// CRTC's framebuffer, so primary plane size should be CRTC size.
 	uint32_t frame_width = primary->md.crtc_w;
@@ -256,7 +279,7 @@ static ssize_t _on_frame_msg(RfStreamer *this)
 
 out:
 	for (size_t i = 0; i < length; ++i)
-		for (int j = 0; j < bufs[i].md.length; ++j)
+		for (unsigned int j = 0; j < bufs[i].md.length; ++j)
 			close(bufs[i].fds[j]);
 
 	if (ret < 0)
@@ -269,8 +292,8 @@ out:
 static ssize_t _on_card_path_msg(RfStreamer *this)
 {
 	g_autofree char *msg = NULL;
-	size_t length = 0;
 	ssize_t ret = 0;
+	size_t length = 0;
 	g_autoptr(GError) error = NULL;
 	GInputStream *is =
 		g_io_stream_get_input_stream(G_IO_STREAM(this->connection));
@@ -327,8 +350,38 @@ out:
 	return ret;
 }
 
-static gboolean
-_on_socket_in(GSocket *socket, GIOCondition condition, gpointer data)
+static ssize_t _on_auth_msg(RfStreamer *this)
+{
+	struct rf_auth auth;
+	size_t length = 0;
+	ssize_t ret = 0;
+	g_autoptr(GError) error = NULL;
+	GInputStream *is =
+		g_io_stream_get_input_stream(G_IO_STREAM(this->connection));
+
+	ret = g_input_stream_read(is, &length, sizeof(length), NULL, &error);
+	if (ret <= 0 || length != 1)
+		goto out;
+
+	ret = g_input_stream_read(is, &auth, sizeof(auth), NULL, &error);
+	if (ret <= 0 || auth.pid < 0)
+		goto out;
+
+	g_debug("Auth: Received auth message for PID %d with result %s.",
+		auth.pid,
+		auth.ok ? "OK" : "not OK");
+	g_signal_emit(this, sigs[SIG_AUTH], 0, auth.pid, auth.ok);
+
+out:
+	if (ret < 0)
+		g_warning(
+			"Auth: Failed to receive auth message: %s.",
+			error->message
+		);
+	return ret;
+}
+
+static int _on_socket_in(GSocket *socket, GIOCondition condition, void *data)
 {
 	RfStreamer *this = data;
 
@@ -360,6 +413,9 @@ _on_socket_in(GSocket *socket, GIOCondition condition, gpointer data)
 	case RF_MSG_TYPE_CONNECTOR_NAME:
 		ret = _on_connector_name_msg(this);
 		break;
+	case RF_MSG_TYPE_AUTH:
+		ret = _on_auth_msg(this);
+		break;
 	default:
 		break;
 	}
@@ -380,7 +436,6 @@ static void _dispose(GObject *o)
 
 	rf_streamer_stop(this);
 	g_clear_object(&this->address);
-	g_clear_object(&this->client);
 
 	G_OBJECT_CLASS(rf_streamer_parent_class)->dispose(o);
 }
@@ -434,12 +489,24 @@ static void rf_streamer_class_init(RfStreamerClass *klass)
 		1,
 		G_TYPE_STRING
 	);
+	sigs[SIG_AUTH] = g_signal_new(
+		"auth",
+		RF_TYPE_STREAMER,
+		0,
+		0,
+		NULL,
+		NULL,
+		NULL,
+		G_TYPE_NONE,
+		2,
+		G_TYPE_INT,
+		G_TYPE_BOOLEAN
+	);
 }
 
 static void rf_streamer_init(RfStreamer *this)
 {
 	this->config = NULL;
-	this->client = NULL;
 	this->address = NULL;
 	this->connection = NULL;
 	this->io_flags = G_IO_IN | G_IO_PRI;
@@ -470,8 +537,6 @@ void rf_streamer_set_socket_path(RfStreamer *this, const char *socket_path)
 	g_return_if_fail(socket_path != NULL);
 
 	g_clear_object(&this->address);
-	g_clear_object(&this->client);
-	this->client = g_socket_client_new();
 	this->address = g_unix_socket_address_new(socket_path);
 }
 
@@ -483,7 +548,7 @@ int rf_streamer_start(RfStreamer *this)
 		return 0;
 
 	this->last_frame_time = -1;
-	unsigned int fps = rf_config_get_fps(this->config);
+	const unsigned int fps = rf_config_get_fps(this->config);
 	this->max_interval = 1000000 / fps;
 	g_message("Frame: Got FPS %u.", fps);
 	this->desktop_width = rf_config_get_desktop_width(this->config);
@@ -507,7 +572,10 @@ int rf_streamer_start(RfStreamer *this)
 
 	g_autoptr(GError) error = NULL;
 	this->connection = g_socket_client_connect(
-		this->client, G_SOCKET_CONNECTABLE(this->address), NULL, &error
+		G_SOCKET_CLIENT(this),
+		G_SOCKET_CONNECTABLE(this->address),
+		NULL,
+		&error
 	);
 	if (this->connection == NULL) {
 		g_warning(
@@ -525,7 +593,7 @@ int rf_streamer_start(RfStreamer *this)
 	_schedule_frame_msg(this);
 
 	this->running = true;
-	g_debug("Emitting ReFrame Streamer start signal.");
+	g_debug("Signal: Emitting ReFrame Streamer start signal.");
 	g_signal_emit(this, sigs[SIG_START], 0);
 	return 0;
 }
@@ -544,27 +612,19 @@ void rf_streamer_stop(RfStreamer *this)
 	if (!this->running)
 		return;
 
-	g_debug("Emitting ReFrame Streamer stop signal.");
+	g_debug("Signal: Emitting ReFrame Streamer stop signal.");
 	g_signal_emit(this, sigs[SIG_STOP], 0);
 	this->running = false;
 
-	if (this->source != NULL) {
+	if (this->source != NULL)
 		g_source_destroy(this->source);
-		g_clear_pointer(&this->source, g_source_unref);
-	}
+	g_clear_pointer(&this->source, g_source_unref);
 	if (this->timer_id != 0) {
 		g_source_remove(this->timer_id);
 		this->timer_id = 0;
 	}
-	g_autoptr(GError) error = NULL;
-	g_io_stream_close(G_IO_STREAM(this->connection), NULL, &error);
-	if (error != NULL) {
-		g_warning(
-			"Failed to close ReFrame Streamer connection: %s.",
-			error->message
-		);
-		return;
-	}
+	// Dropping the last reference of it will automatically close IO
+	// streams and socket.
 	g_clear_object(&this->connection);
 }
 
@@ -620,6 +680,9 @@ void rf_streamer_send_pointer_event(
 	const uint32_t desktop_height =
 		this->desktop_height > 0 ? this->desktop_height :
 					   this->monitor_y + this->frame_height;
+	// This may happen if we are still not getting the first frame.
+	if (desktop_width == 0 || desktop_height == 0)
+		return;
 	// Typically desktop environment will map uinput `EV_ABS` max size to
 	// the whole virtual desktop, so we need to convert the position to
 	// global position in the virtual desktop.
@@ -691,3 +754,11 @@ void rf_streamer_send_pointer_event(
 
 	_send_input_msg(this, ies, length);
 }
+
+void rf_streamer_auth(RfStreamer *this, pid_t pid)
+{
+	g_return_if_fail(RF_IS_STREAMER(this));
+	g_return_if_fail(pid >= 0);
+
+	_send_auth_msg(this, pid);
+}
