@@ -1,5 +1,6 @@
 #include <stdint.h>
 #include <stdbool.h>
+#include <locale.h>
 #include <glib.h>
 #include <glib-unix.h>
 #include <gio/gio.h>
@@ -12,7 +13,6 @@
 #include "config.h"
 #include "rf-common.h"
 #include "rf-config.h"
-#include "rf-buffer.h"
 
 #ifdef HAVE_LIBSYSTEMD
 #	include <systemd/sd-daemon.h>
@@ -42,7 +42,7 @@
 
 #define _write_may(fd, buf, count)                                                                              \
 	G_STMT_START {                                                                                          \
-		ssize_t e = write((fd), (buf), (count));                                                        \
+		size_t e = write((fd), (buf), (count));                                                        \
 		if (e != (count))                                                                               \
 			g_warning(                                                                              \
 				"Input: Failed to write %ld bytes to %d at line %d, actually wrote %ld bytes.", \
@@ -66,8 +66,93 @@ struct _this {
 	uint32_t cursor_id;
 	int ufd;
 	bool wakeup;
+	bool skip_auth;
 };
 
+static int _auth_pid(struct _this *this, pid_t pid, const char *target)
+{
+	if (this->skip_auth)
+		return 0;
+
+	g_autoptr(GError) error = NULL;
+	g_autofree char *proc_exe = g_strdup_printf(
+		G_DIR_SEPARATOR_S "proc" G_DIR_SEPARATOR_S
+				  "%d" G_DIR_SEPARATOR_S "exe",
+		pid
+	);
+	g_autofree char *bin = g_file_read_link(proc_exe, &error);
+	if (bin == NULL) {
+		g_warning(
+			"Auth: Failed to read link of %s: %s.",
+			proc_exe,
+			error->message
+		);
+		return -1;
+	}
+	g_debug("Auth: Authenticating process executable binary %s.", bin);
+	if (g_strcmp0(bin, target) != 0)
+		return -2;
+	return 0;
+}
+
+static ssize_t _send_auth_msg(struct _this *this, pid_t pid, bool ok)
+{
+	ssize_t ret = 0;
+	g_autoptr(GError) error = NULL;
+	GOutputStream *os =
+		g_io_stream_get_output_stream(G_IO_STREAM(this->connection));
+
+	ret = rf_send_header(this->connection, RF_MSG_TYPE_AUTH, 1, &error);
+	if (ret <= 0)
+		goto out;
+	struct rf_auth auth;
+	auth.pid = pid;
+	auth.ok = ok;
+	ret = g_output_stream_write(os, &auth, sizeof(auth), NULL, &error);
+
+out:
+	if (ret < 0)
+		g_warning(
+			"Auth: Failed to send auth message: %s.", error->message
+		);
+	else if (ret > 0)
+		g_debug("Auth: Sent auth message for PID %d with result %s.",
+			auth.pid,
+			auth.ok ? "OK" : "not OK");
+	return ret;
+}
+
+static ssize_t _on_auth_msg(struct _this *this)
+{
+	ssize_t ret = 0;
+	size_t length = 0;
+	g_autoptr(GError) error = NULL;
+	GInputStream *is =
+		g_io_stream_get_input_stream(G_IO_STREAM(this->connection));
+
+	ret = g_input_stream_read(is, &length, sizeof(length), NULL, &error);
+	if (ret <= 0 || length != 1)
+		goto out;
+	pid_t pid = -1;
+	ret = g_input_stream_read(is, &pid, sizeof(pid), NULL, &error);
+	if (ret <= 0 || pid < 0)
+		goto out;
+
+	g_debug("Auth: Received auth message for PID %d.", pid);
+	bool ok = _auth_pid(
+			  this, pid, BINDIR G_DIR_SEPARATOR_S "reframe-session"
+		  ) == 0;
+	return _send_auth_msg(this, pid, ok);
+
+out:
+	if (ret < 0)
+		g_warning(
+			"Auth: Failed to receive auth message: %s.",
+			error->message
+		);
+	return ret;
+}
+
 // You need to explicitly cast the type of returned value.
 static uint64_t _get_plane_prop(
 	int cfd,
@@ -81,7 +166,7 @@ static uint64_t _get_plane_prop(
 		cfd, plane_id, DRM_MODE_OBJECT_PLANE
 	);
 	if (props == NULL)
-		return -1;
+		return value;
 	for (size_t i = 0; i < props->count_props; ++i) {
 		drmModePropertyRes *prop =
 			drmModeGetProperty(cfd, props->props[i]);
@@ -130,7 +215,7 @@ static uint32_t _get_plane_id(int cfd, uint32_t crtc_id, uint32_t type)
 	return plane_id;
 }
 
-static int _export_fb2(int cfd, RfBuffer *b, uint32_t fb_id)
+static int _export_fb2(int cfd, struct rf_buffer *b, uint32_t fb_id)
 {
 	drmModeFB2 *fb = drmModeGetFB2(cfd, fb_id);
 	if (fb == NULL)
@@ -150,7 +235,7 @@ static int _export_fb2(int cfd, RfBuffer *b, uint32_t fb_id)
 	b->md.modifier = fb->flags & DRM_MODE_FB_MODIFIERS ?
 				 fb->modifier :
 				 DRM_FORMAT_MOD_INVALID;
-	for (int i = 0; i < b->md.length; ++i) {
+	for (unsigned int i = 0; i < b->md.length; ++i) {
 		b->md.offsets[i] = fb->offsets[i];
 		b->md.pitches[i] = fb->pitches[i];
 	}
@@ -158,7 +243,7 @@ static int _export_fb2(int cfd, RfBuffer *b, uint32_t fb_id)
 	return b->md.length;
 }
 
-static int _export_fb(int cfd, RfBuffer *b, uint32_t fb_id)
+static int _export_fb(int cfd, struct rf_buffer *b, uint32_t fb_id)
 {
 	drmModeFB *fb = drmModeGetFB(cfd, fb_id);
 	if (fb == NULL)
@@ -180,12 +265,13 @@ static int _export_fb(int cfd, RfBuffer *b, uint32_t fb_id)
 	return b->md.length;
 }
 
-static int _make_buffer(int cfd, RfBuffer *b, uint32_t plane_id, uint32_t type)
+static int
+_make_buffer(int cfd, struct rf_buffer *b, uint32_t plane_id, uint32_t type)
 {
 	drmModePlane *plane = drmModeGetPlane(cfd, plane_id);
 	if (plane == NULL)
 		return 0;
-	uint32_t fb_id = plane->fb_id;
+	const uint32_t fb_id = plane->fb_id;
 	drmModeFreePlane(plane);
 	if (fb_id == 0)
 		return 0;
@@ -228,7 +314,7 @@ static int _make_buffer(int cfd, RfBuffer *b, uint32_t plane_id, uint32_t type)
 }
 
 static ssize_t
-_send_buffer(GSocketConnection *connection, RfBuffer *b, GError **error)
+_send_buffer(GSocketConnection *connection, struct rf_buffer *b, GError **error)
 {
 	ssize_t ret = 0;
 	GOutputVector iov = { &b->md, sizeof(b->md) };
@@ -236,7 +322,7 @@ _send_buffer(GSocketConnection *connection, RfBuffer *b, GError **error)
 	// This won't take the ownership so we need to close fds.
 	//
 	// See <https://docs.gtk.org/gio/method.UnixFDList.append.html>.
-	for (int i = 0; i < b->md.length; ++i)
+	for (unsigned int i = 0; i < b->md.length; ++i)
 		g_unix_fd_list_append(fds, b->fds[i], NULL);
 	// This won't take the ownership so we need to free GUnixFDList.
 	GSocketControlMessage *msg = g_unix_fd_message_new_with_fd_list(fds);
@@ -244,13 +330,13 @@ _send_buffer(GSocketConnection *connection, RfBuffer *b, GError **error)
 	ret = g_socket_send_message(
 		socket, NULL, &iov, 1, &msg, 1, G_SOCKET_MSG_NONE, NULL, error
 	);
-	g_object_unref(fds);
-	g_object_unref(msg);
+	g_clear_object(&fds);
+	g_clear_object(&msg);
 	return ret;
 }
 
 static ssize_t
-_send_frame_msg(struct _this *this, size_t length, RfBuffer *bufs)
+_send_frame_msg(struct _this *this, size_t length, struct rf_buffer *bufs)
 {
 	ssize_t ret = 0;
 	g_autoptr(GError) error = NULL;
@@ -276,9 +362,9 @@ static ssize_t _on_frame_msg(struct _this *this)
 {
 	g_debug("Frame: Received frame message.");
 
-	RfBuffer bufs[RF_MAX_BUFS];
-	size_t length = 0;
+	struct rf_buffer bufs[RF_MAX_BUFS];
 	ssize_t ret = 0;
+	size_t length = 0;
 	g_autoptr(GError) error = NULL;
 	GInputStream *is =
 		g_io_stream_get_input_stream(G_IO_STREAM(this->connection));
@@ -326,7 +412,7 @@ static ssize_t _on_frame_msg(struct _this *this)
 	ret = _send_frame_msg(this, length, bufs);
 
 	for (size_t i = 0; i < length; ++i)
-		for (int j = 0; j < bufs[i].md.length; ++j)
+		for (unsigned int j = 0; j < bufs[i].md.length; ++j)
 			close(bufs[i].fds[j]);
 	return ret;
 }
@@ -336,8 +422,8 @@ static ssize_t _on_input_msg(struct _this *this)
 	g_debug("Input: Received input message.");
 
 	g_autofree struct input_event *ies = NULL;
-	size_t length = 0;
 	ssize_t ret = 0;
+	size_t length = 0;
 	g_autoptr(GError) error = NULL;
 	GInputStream *is =
 		g_io_stream_get_input_stream(G_IO_STREAM(this->connection));
@@ -367,8 +453,8 @@ out:
 
 static ssize_t _send_card_path_msg(struct _this *this, const char *card_path)
 {
-	size_t length = strlen(card_path) + 1;
 	ssize_t ret = 0;
+	size_t length = strlen(card_path) + 1;
 	g_autoptr(GError) error = NULL;
 	GOutputStream *os =
 		g_io_stream_get_output_stream(G_IO_STREAM(this->connection));
@@ -389,8 +475,8 @@ out:
 static ssize_t
 _send_connector_name_msg(struct _this *this, const char *connector_name)
 {
-	size_t length = strlen(connector_name) + 1;
 	ssize_t ret = 0;
+	size_t length = strlen(connector_name) + 1;
 	g_autoptr(GError) error = NULL;
 	GOutputStream *os =
 		g_io_stream_get_output_stream(G_IO_STREAM(this->connection));
@@ -461,8 +547,9 @@ _get_connected_card_and_connector(struct _this *this, const char *connector_name
 	while ((name = g_dir_read_name(dir)) != NULL) {
 		if (!g_str_has_prefix(name, "card"))
 			continue;
-		g_autofree char *card_path =
-			g_strdup_printf("/dev/dri/%s", name);
+		g_autofree char *card_path = g_build_filename(
+			G_DIR_SEPARATOR_S "dev", "dri", name, NULL
+		);
 		int cfd = open(card_path, O_RDONLY | O_CLOEXEC);
 		if (cfd < 0)
 			continue;
@@ -666,15 +753,23 @@ static void _clean_uinput(struct _this *this)
 	}
 }
 
+static void _on_sigint(int sig)
+{
+	// Non-zero hints that we didn't clean up.
+	exit(2);
+}
+
 int main(int argc, char *argv[])
 {
+	setlocale(LC_ALL, "");
+
 	g_autofree char *config_path = NULL;
 	g_autofree char *socket_path = NULL;
 	// `gboolean` is `int`, but `bool` may be `char`! Passing `bool` pointer
 	// to `GOptionContext` leads into overflow!
-	gboolean keep_listen = FALSE;
-	gboolean version = FALSE;
-	g_autofree char **args = g_strdupv(argv);
+	int keep_listen = false;
+	int skip_auth = false;
+	int version = false;
 	g_autoptr(GError) error = NULL;
 
 	GOptionEntry options[] = {
@@ -706,6 +801,13 @@ int main(int argc, char *argv[])
 		  &keep_listen,
 		  "Keep listening to socket after disconnection (debug purpose).",
 		  NULL },
+		{ "skip-auth",
+		  'A',
+		  G_OPTION_FLAG_NONE,
+		  G_OPTION_ARG_NONE,
+		  &skip_auth,
+		  "Skip socket client authenticating by always OK (debug purpose).",
+		  NULL },
 		{ NULL,
 		  0,
 		  G_OPTION_FLAG_NONE,
@@ -717,32 +819,47 @@ int main(int argc, char *argv[])
 	g_autoptr(GOptionContext)
 		context = g_option_context_new(" - ReFrame Streamer");
 	g_option_context_add_main_entries(context, options, NULL);
-	if (!g_option_context_parse_strv(context, &args, &error)) {
+	if (!g_option_context_parse(context, &argc, &argv, &error)) {
 		g_warning("Failed to parse options: %s.", error->message);
 		g_clear_pointer(&error, g_error_free);
 	}
+
 	if (version) {
 		g_print(PROJECT_VERSION "\n");
 		return 0;
 	}
-	// Use `g_strdup()` here to make `g_autofree` happy.
-	if (socket_path == NULL)
-		socket_path = g_strdup("/tmp/reframe.sock");
+
+	// We ensure the default dir, user ensure the argument dir.
+	if (socket_path == NULL) {
+		g_mkdir("/tmp/reframe", 0755);
+		rf_set_group("/tmp/reframe");
+		socket_path = g_strdup("/tmp/reframe/reframe.sock");
+	}
+
+	g_message(
+		"Keep listening mode is %s.",
+		keep_listen ? "enabled" : "disabled"
+	);
+	g_message(
+		"Skip authenticating mode is %s.",
+		skip_auth ? "enabled" : "disabled"
+	);
+	g_message("Using configuration file %s.", config_path);
+	g_message("Using socket %s.", socket_path);
 
 	g_autofree struct _this *this = g_malloc0(sizeof(*this));
-	g_message("Using configuration file %s.", config_path);
-	this->config = rf_config_new(config_path);
 	this->cfd = -1;
 	this->ufd = -1;
+	this->skip_auth = skip_auth;
+	this->config = rf_config_new(config_path);
 
 	g_autoptr(GSocketListener) listener = g_socket_listener_new();
-	g_message("Using socket %s.", socket_path);
 
 #ifdef HAVE_LIBSYSTEMD
 	if (sd_listen_fds(0) != 0) {
 		// systemd socket.
 		// We only handle 1 socket.
-		int sfd = SD_LISTEN_FDS_START;
+		const int sfd = SD_LISTEN_FDS_START;
 		g_autoptr(GSocket) socket = g_socket_new_from_fd(sfd, &error);
 		if (error != NULL)
 			g_error("Failed to create socket from systemd fd: %s.",
@@ -750,10 +867,10 @@ int main(int argc, char *argv[])
 		g_socket_listener_add_socket(listener, socket, NULL, &error);
 	} else {
 #endif
-		g_remove(socket_path);
 		// Non-systemd socket.
 		g_autoptr(GSocketAddress)
 			address = g_unix_socket_address_new(socket_path);
+		g_remove(socket_path);
 		g_socket_listener_add_address(
 			listener,
 			address,
@@ -763,16 +880,15 @@ int main(int argc, char *argv[])
 			NULL,
 			&error
 		);
+		rf_set_group(socket_path);
+		g_chmod(socket_path, 0660);
 #ifdef HAVE_LIBSYSTEMD
 	}
 #endif
 	if (error != NULL)
 		g_error("Failed to listen to socket: %s.", error->message);
 
-	g_message(
-		"Keep listening mode is %s.",
-		keep_listen ? "enabled" : "disabled"
-	);
+	signal(SIGINT, _on_sigint);
 	do {
 		this->connection =
 			g_socket_listener_accept(listener, NULL, NULL, &error);
@@ -780,6 +896,16 @@ int main(int argc, char *argv[])
 			g_error("Failed to accept connection: %s.",
 				error->message);
 
+		GSocket *socket =
+			g_socket_connection_get_socket(this->connection);
+		const pid_t pid = rf_get_socket_pid(socket);
+		if (_auth_pid(
+			    this, pid, BINDIR G_DIR_SEPARATOR_S "reframe-server"
+		    ) != 0) {
+			g_warning("Got disallowed socket client PID %d.", pid);
+			goto close;
+		}
+
 		g_message("ReFrame Server connected.");
 
 		_setup_uinput(this);
@@ -810,6 +936,9 @@ int main(int argc, char *argv[])
 			case RF_MSG_TYPE_INPUT:
 				ret = _on_input_msg(this);
 				break;
+			case RF_MSG_TYPE_AUTH:
+				ret = _on_auth_msg(this);
+				break;
 			default:
 				break;
 			}
@@ -822,15 +951,14 @@ int main(int argc, char *argv[])
 		_clean_drm(this);
 		_clean_uinput(this);
 
-		g_io_stream_close(G_IO_STREAM(this->connection), NULL, &error);
-		if (error != NULL)
-			g_error("Failed to close socket connection: %s.",
-				error->message);
+	close:
+		// Dropping the last reference of it will automatically close IO
+		// streams and socket.
 		g_clear_object(&this->connection);
 	} while (keep_listen);
 
 	g_socket_listener_close(listener);
-	g_object_unref(this->config);
+	g_clear_object(&this->config);
 
 	return 0;
 }
